#include "argumentc.h"
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>

int Argumentc_Create(Argumentc* target, int argc, const char** argv)
{
	if (target == NULL) return EDESTADDRREQ;
	if (argv == NULL) return EINVAL;

	int array_code = DynamicArray_Create(
		&target->array,
		sizeof(Option),
		argc * 2,
		NULL
	);
	if (array_code != EXIT_SUCCESS) return ENOMEM;

	int i;
	for (i = 0; i < argc; i++) {
		StringView arg = StringView_FromString(argv[i]);
		enum OptionType type;
		Option opt;

		// Determine Option type
		switch (arg.length) {
		case 0:
		case 1:
			type = OPTIONTYPE_ARGUMENT;
			break;
		case 2:
			if (StringView_Equal(
				arg,
				StringView_FromString("--"))) {
				type = OPTIONTYPE_DELIM;
			} else if (StringView_StartsWith(
				arg,
				StringView_FromString("-"))) {
				type = OPTIONTYPE_SHORT;
			} else {
				type = OPTIONTYPE_ARGUMENT;
			}
			break;
		default:
			if (StringView_StartsWith(
				arg,
				StringView_FromString("--"))) {
				type = OPTIONTYPE_LONG;

			} else if (StringView_StartsWith(
				arg,
				StringView_FromString("-"))) {
				type = OPTIONTYPE_SHORT;
			} else {
				type = OPTIONTYPE_ARGUMENT;
			}
			break;
		}

		opt.type = type;

		switch (type) {
		case OPTIONTYPE_NONE:
		case OPTIONTYPE_ARGUMENT:
		case OPTIONTYPE_DELIM:
			opt.content = arg;
			if (DynamicArray_Append(&target->array, &opt))
				goto defer_arraylist;
			break;
		case OPTIONTYPE_LONG:
			opt.content = StringView_Slice(
				arg,
				2,
				arg.length);
			if (DynamicArray_Append(&target->array, &opt))
				goto defer_arraylist;
			break;
		case OPTIONTYPE_SHORT:
			arg.length -= 1;
			arg.source++;
			while (arg.length) {
				opt.content = StringView_Slice(arg, 0, 1);
				if (DynamicArray_Append(&target->array, &opt))
					goto defer_arraylist;
				arg.length--;
				arg.source++;
			}
		}

		if (opt.type == OPTIONTYPE_DELIM) break;
	}

	for (++i; i < argc; i++) {
		Option opt;
		opt.type = OPTIONTYPE_ARGUMENT;
		opt.content = StringView_FromString(argv[i]);
		if (DynamicArray_Append(&target->array, &opt))
			goto defer_arraylist;
	}

	return EXIT_SUCCESS;
defer_arraylist:
	DynamicArray_Destroy(&target->array);
	return ENOMEM;
}

void Argumentc_Destroy(Argumentc* argumentc)
{
	DynamicArray_Destroy(&argumentc->array);
	return;
}

bool Argumentc_HaveNextOption(Argumentc* argumentc)
{
	return DynamicArray_GetLength(&argumentc->array) != 0;
}

static Option _Argumentc_PopOption(Argumentc* argumentc, size_t index)
{
	Option* next = DynamicArray_GetPointer(&argumentc->array, index);
	Option local = *next;
	DynamicArray_Remove(&argumentc->array, index);

	return local;
}

Option Argumentc_PopNextOption(Argumentc* argumentc)
{
	return _Argumentc_PopOption(argumentc, 0);
}

static size_t _Argumentc_FindOption(Argumentc* argumentc, Option option)
{
	for (size_t i = 0; i < DynamicArray_GetLength(&argumentc->array); i++) {
		Option* array_option = DynamicArray_GetPointer(&argumentc->array, i);

		if (array_option->type != option.type) continue;
		if (StringView_Equal(array_option->content, option.content)) {
			return i;
		}
	}

	return SIZE_MAX;
}

bool Argumentc_HaveOption(Argumentc* argumentc, StringView name)
{
	for (size_t i = 0; i < DynamicArray_GetLength(&argumentc->array); i++) {
		Option* array_option = DynamicArray_GetPointer(&argumentc->array, i);

		if (StringView_Equal(array_option->content, name)) {
			return true;
		}
	}

	return false;
}

/*
static size_t _Argumentc_FindByType(Argumentc* argumentc, enum OptionType type, size_t start)
{
	for (size_t i = start; i < DynamicArray_GetLength(&argumentc->array); i++) {
		Option* array_option = DynamicArray_GetPointer(&argumentc->array, i);

		if (array_option->type == type) {
			return i;
		}
	}

	return SIZE_MAX;
}
*/

Option Argumentc_PopShortOption(Argumentc* argumentc, StringView name)
{
	Option short_option = { OPTIONTYPE_SHORT, name};
	size_t index = _Argumentc_FindOption(argumentc, short_option);

	if (index == SIZE_MAX) {
		return (Option) {OPTIONTYPE_NONE, name};
	}

	return _Argumentc_PopOption(argumentc, index);
}

Option Argumentc_PopLongOption(Argumentc* argumentc, StringView name)
{
	Option long_option = { OPTIONTYPE_LONG, name};
	size_t index = _Argumentc_FindOption(argumentc, long_option);

	if (index == SIZE_MAX) {
		return (Option) {OPTIONTYPE_NONE, name};
	}

	return _Argumentc_PopOption(argumentc, index);
}

OptionArgument Argumentc_PopShortArgument(Argumentc* argumentc, StringView name)
{
	Option find_option = { OPTIONTYPE_SHORT, name};
	size_t index = _Argumentc_FindOption(argumentc, find_option);

	if (index == SIZE_MAX) {
		return (OptionArgument) {{OPTIONTYPE_NONE, name}, {OPTIONTYPE_NONE, name}};
	}

	Option* short_option = DynamicArray_GetPointer(&argumentc->array, index);

	size_t argument_index = index;
	Option* array_current = DynamicArray_GetPointer(&argumentc->array, argument_index);
	while (array_current->type == OPTIONTYPE_SHORT) {
		argument_index++;
		array_current++;
	}

	if (array_current->type != OPTIONTYPE_ARGUMENT) {
		return (OptionArgument) {{OPTIONTYPE_NONE, name}, {OPTIONTYPE_NONE, name}};
	}

	OptionArgument pair;
	pair.argument = *array_current;
	pair.option   = *short_option;

	DynamicArray_Remove(&argumentc->array, argument_index);
	DynamicArray_Remove(&argumentc->array, index);

	return pair;
}

OptionArgument Argumentc_PopLongArgument(Argumentc* argumentc, StringView name)
{
	Option long_option = {OPTIONTYPE_LONG, name};
	size_t index = _Argumentc_FindOption(argumentc, long_option);

	if (index == SIZE_MAX) {
		return (OptionArgument) {{OPTIONTYPE_NONE, name}, {OPTIONTYPE_NONE, name}};
	}

	Option* array_long = DynamicArray_GetPointer(&argumentc->array, index);
	Option* array_arg  = DynamicArray_GetPointer(&argumentc->array, index + 1);

	if (array_arg == NULL || array_arg->type != OPTIONTYPE_ARGUMENT) {
		return (OptionArgument) {{OPTIONTYPE_NONE, name}, {OPTIONTYPE_NONE, name}};
	}

	OptionArgument pair;
	pair.argument = *array_arg;
	pair.option   = *array_long;

	DynamicArray_Remove(&argumentc->array, index); // remove option
	DynamicArray_Remove(&argumentc->array, index); // remove argument

	return pair;
}

bool Argumentc_PopLongOptionSwitch(Argumentc* argumentc, StringView name, bool bdefault)
{
	Option switch_option = Argumentc_PopLongOption(argumentc, name);

	return switch_option.type == OPTIONTYPE_LONG ? true : bdefault;
}

bool Argumentc_PopShortOptionSwitch(Argumentc* argumentc, StringView name, bool bdefault)
{
	Option switch_option = Argumentc_PopShortOption(argumentc, name);

	return switch_option.type == OPTIONTYPE_SHORT ? true : bdefault;
}

int Argumentc_PopLongOptionInt(Argumentc* argumentc, StringView name, int idefault)
{
	OptionArgument int_option = Argumentc_PopLongArgument(argumentc, name);

	return int_option.option.type == OPTIONTYPE_LONG ?
		StringView_ParseInt(int_option.argument.content) :
		idefault;
}

int Argumentc_PopShortOptionInt(Argumentc* argumentc, StringView name, int idefault)
{
	OptionArgument int_option = Argumentc_PopShortArgument(argumentc, name);

	return int_option.option.type == OPTIONTYPE_SHORT ?
		StringView_ParseInt(int_option.argument.content) :
		idefault;
}

const char* OptionType_ToString(enum OptionType type)
{
	switch (type) {
		case OPTIONTYPE_NONE:
			return "OPTIONTYPE_NONE";
		case OPTIONTYPE_SHORT:
			return "OPTIONTYPE_SHORT";
		case OPTIONTYPE_LONG:
			return "OPTIONTYPE_LONG";
		case OPTIONTYPE_DELIM:
			return "OPTIONTYPE_DELIM";
		case OPTIONTYPE_ARGUMENT:
			return "OPTIONTYPE_ARGUMENT";
		default:
			return "OPTIONTYPE_INVALID";
	}
}
